#!/bin/ksh
#
# $Id: flashdist.sh,v 1.9 2004/01/21 06:56:50 blovett Exp $
#
# Copyright (c) 2002, 2003 Chris Cappuccio.
# Modified by Ron Rosson.
#
# This quick and dirty script will setup a flash media with an OpenBSD/i386
# distribution specified by <distfile>.  This script makes many assumptions,
# but should work just dandy with most any OpenBSD/i386 3.x system.  I've
# used it with USB CompactFlash and I suspect it will work fine for IDE CF
# or DOM or anything perhaps.   Maybe.
#
# This script contains a stupid method occasionally works to make this
# media bootable on a destination which uses a different c/h/s translation
# than the host system.  Of course, this is really just a hack.
#
# To use this script:
#  1. Run it without using any modified values.  Your system should
#     be able to determine the correct geometry for the media you are
#     installing to.  If the media already has a failed OpenBSD installation
#     on it, you can igore the disklabel's values by using -d.
#  2. If this fails to boot on the destination system with 'Bad magic',
#     or other boot errors, then the destination system is not using the
#     same geometry that your install system detected.  You need to manually
#     specify a geometry which will work.  Try the sample ones in here.
#  3. If the destination system now boots when you force a geometry,
#     then you can use disklabel -d wd0 (or sd0 or whatever your boot
#     device is) once you are booted to find out the real geometry
#     that you should be using.
#
# If you cannot boot the resulting media initialized by this script,
# you may need to fill in the following values with the output from
# disklabel -d <device> on the destination system  Of course, you also need
# to uncomment the values (remove the pound sign, eh?)
#
# If you can't get the disklabel -d values for the media on your destination
# system, you can either ask someone else who has the same setup to do
# disklabel -d and send you the output or find some other creative way, like
# writing the miniroot image to your media or something...
#
# In any event, the real goal here is to get something that is bootable
# so you can run disklabel -d wd0 when you are booted on the embedded
# device and get the real values that you should be using.  That is,
# you don't need to use the absolutely correct values for the first run around!

umask 022

vers="20030122"

PATH=/usr/bin:/usr/sbin:/sbin:/bin

usage()
{
 echo "Usage: flashdist.sh [-d] <flash size> <disk device name> <distfile> <bsd kernel> <openbsd base>"
 echo 
 echo "-d >> this option will cause disklabel to always initialize the media with"
 echo "      the kernel's current idea of the disk geometry even if the old disklabel"
 echo "      has a different idea of the geometry"
 echo "flash size >> The size of the CF you want to use in Megs"
 echo "disk device name >> the base name of your flash media, such as sd0"
 echo "distfile >> distribution list file name"
 echo "bsd kernel >> kernel to install on your flash"
 echo "openbsd base >> unpacked OpenBSD distribution (unpacked baseXX.tgz, etcXX.tgz)"
 echo
 echo "example:"
 echo "flashdist.sh 64 sd0 flashsmall.txt /usr/src/sys/arch/i386/compile/NET4501/bsd /tmp/openbsd"
 exit
}

echo flashdist.sh $vers chris@nmedia.net modded by insane@oneinsane.net

if [ "$1" == "-d" ]; then
 dflag="-d"
 shift
else
 dflag=""
fi

# (The sample values which come with this script are for the SanDisk SDCFB-64
# as seen by the Soekris net4501.)

if [ "$1" == "" ]; then
    usage
else
    case $1 in 
	32)
	    # 32Meg CF
	    totalsize=62720         # "total sectors:"
	    bytessec=512            # "bytes/sector:"
	    sectorstrack=32         # "sectors/track:"
	    sectorscylinder=256     # "sectors/cylinder:"
	    trackscylinder=8        # "tracks/cylinder:"
	    cylinders=245           # "cylinders:"
	    ;;
	64)
	    # SanDisk SDCFB-64
	    totalsize=125440        # "total sectors:"
	    bytessec=512            # "bytes/sector:"
	    sectorstrack=32         # "sectors/track:"
	    sectorscylinder=256     # "sectors/cylinder:"
	    trackscylinder=8        # "tracks/cylinder:"
	    cylinders=490           # "cylinders:"
	    ;;
	128)
	    # 128Meg CF
	    totalsize=250368	    # "total sectors:" 
	    bytessec=512	    # "bytes/sector:"
	    sectorstrack=32	    # "sectors/track:"
	    sectorscylinder=2048    # "sectors/cylinder:"
	    trackscylinder=64	    # "tracks/cylinder:"
	    cylinders=122	    # "cylinders:"
	    ;;
	256)
	    # 256Meg CF
	    totalsize=501760        # "total sectors:"
	    bytessec=512            # "bytes/sector:"
	    sectorstrack=32         # "sectors/track:"
	    sectorscylinder=512     # "sectors/cylinder:"
	    trackscylinder=16       # "tracks/cylinder:"
	    cylinders=980           # "cylinders:"
	    ;;
	1024)
	    # 1gb IBM Microdrive
	    totalsize=2104704	    # "total sectors:" 
	    bytessec=512	    # "bytes/sector:"
	    sectorstrack=63	    # "sectors/track:"
	    sectorscylinder=4032    # "sectors/cylinder:"
	    trackscylinder=64	    # "tracks/cylinder:"
	    cylinders=522	    # "cylinders:"
	    ;;
	*)
	    echo "Please edit this file and flashdist.sh and send"
	    echo "the diffs to the OpenSoekris Team."
	    exit 0
	    ;;
    esac
fi

if [ "$2" == "" ]; then
 usage
else
 disk=/dev/$2c
 if [ ! -b $disk ]; then
   echo $1 is an invalid disk device, $disk not found or not block device
   exit
 else
  echo Using disk device: $2
 fi
fi

if [ "$3" == "" ]; then
 usage
else
 if [ ! -f "$3" ]; then
   echo $3 is an invalid distfile, file not found
   exit
 else
   echo Using distfile: $3
 fi
fi

if [ "$4" == "" ]; then
 usage
else
 if [ ! -f "$4" ]; then
   echo $4 is an invalid kernel, file not found
   exit
 else
   echo Copying kernel from: $4
 fi
fi

if [ "$5" == "" ]; then
 usage
else
 if [ ! -d "$5" ]; then
   echo $5 is not a directory
   exit
 fi
fi

echo
echo Please pay attention to any error messages that you may
echo receive from the commands this script it using.  If you
echo end up having problems, they could explain why.
echo
echo WARNING: This will erase ALL DATA on the $2 disk device!
echo -n Press enter key to continue or Control-C to abort...
read whatever
echo

# XXX somewhat i386 specific for now...
#

TA=`mktemp /tmp/_aaaXXXXXX`
TB=`mktemp /tmp/_bbbXXXXXX`

echo Setting up disklabel...
if ! disklabel $dflag $2 > $TA; then
  echo "Disklabel failed: disklabel $dflag $2 > $TA"
  echo Please fix something.  Aborting...
  echo
  exit
fi
echo

checkd()
{
 if [ "$dflag" == "-d" ]; then
  echo You cannot specify geometry manually and specify the -d flag
  echo "(Please do not specify geometry with the -d switch)"
  echo
  exit
 fi
}

if [ "$totalsize" == "" ]; then
 totalsize=`cat $TA | egrep ^"total sectors:" | awk ' {print $3} '`
else
 checkd
 x=`cat $TA | egrep ^"total sectors:" | awk ' {print $3} '`
 sed -e "s/total sectors: $x/total sectors: $totalsize/" < $TA >$TB
 mv $TB $TA
fi
if [ "$bytessec" == "" ]; then
 bytessec=`cat $TA | egrep ^"bytes/sector:" | awk ' {print $2} '`
else
 checkd
 x=`cat $TA | egrep ^"bytes/sector:" | awk ' {print $2} '`
 sed -e "s/bytes\\/sector: $x/bytes\\/sector: $bytessec/" < $TA >$TB
 mv $TB $TA
fi
if [ "$sectorstrack" == "" ]; then
 sectorstrack=`cat $TA | egrep ^"sectors/track:" | awk ' {print $2} '`
else
 checkd
 x=`cat $TA | egrep ^"sectors/track:" | awk ' {print $2} '`
 sed -e "s/sectors\\/track: $x/sectors\\/track: $sectorstrack/" < $TA >$TB
 mv $TB $TA
fi
if [ "$sectorscylinder" == "" ]; then
 sectorscylinder=`cat $TA | egrep ^"sectors/cylinder:" | awk ' {print $2} '`
else
 checkd
 x=`cat $TA | egrep ^"sectors/cylinder:" | awk ' {print $2} '`
 sed -e "s/sectors\\/cylinder: $x/sectors\\/cylinder: $sectorscylinder/" < $TA >$TB
 mv $TB $TA
fi
if [ "$trackscylinder" == "" ]; then
 trackscylinder=`cat $TA | egrep ^"tracks/cylinder:" | awk ' {print $2} '`
else
 checkd
 x=`cat $TA | egrep ^"tracks/cylinder:" | awk ' {print $2} '`
 sed -e "s/tracks\\/cylinder: $x/tracks\\/cylinder: $trackscylinder/" < $TA >$TB
 mv $TB $TA
fi
if [ "$cylinders" == "" ]; then
 cylinders=`cat $TA | egrep ^"cylinders:" | awk ' {print $2} '`
else
 checkd
 x=`cat $TA | egrep ^"cylinders:" | awk ' {print $2} '`
 sed -e "s/cylinders: $x/cylinders: $cylinders/" < $TA >$TB
 mv $TB $TA
fi

let totalbytes=$totalsize*$bytessec 

echo "The install script is using the following parameters:"
echo "Total size of media: $totalsize sectors ($totalbytes bytes)"
echo "Bytes/Sector: $bytessec"
echo "Sectors/Track: $sectorstrack"
echo "Sectors/Cylinder: $sectorscylinder"
echo "Tracks/Cylinder (heads): $trackscylinder"
echo "Cylinders: $cylinders"
echo
echo -n Press enter key to continue or Control-C to abort...
read whatever
echo

notfound=""

echo Checking distribution list...

# Peanut Butter Wolf, give the people some technology!
for i in `cat $3`; do
 if [ ! -f $5/$i ]; then
  echo $i NOT FOUND
  notfound=1
 fi
done

if [ "$notfound" == "1" ]; then
  echo You are missing files from your chosen distribution list.
  echo All of the file names specified in $3 must exist in the
  echo $5 directory.
  exit
fi

echo

echo Updating MBR and partition table...

fdisk -c $cylinders -h $trackscylinder -s $sectorstrack -f $5/usr/mdec/mbr -e $2 << __EOC >/dev/null
reinit
update
write
quit
__EOC

echo "Note, you may ignore \"sysctl(machdep.bios.diskinfo)\" errors if present."
echo

# This $asize method seems to create a disklabel which works
# under other c/h/s translations then the one we installed on.  This probably
# is complete snake oil.
#
# It causes the partition to end on a cylinder boundary.  The idea here is that
# 1. different c/h/s translations on our destinations (like IDE CF or DOM) will
# have a higher number of cylinders then our host USB CF system, and
# 2. our cylinder boundaries will lie on the same sectors when you compare
# the host cylinder boundary sector numbers with the destination cylinder
# boundary sector numbers.  All bets are off if we are 1. going from a host
# system that come up with significantly different numbers of sectors then the
# destination system, 2. systems that use different numbers of bytes per sector,
# or 3. if we are going from a host system with a high cylinder count to a
# destination system that uses a low cylinder count to represent the media.
#
# Of course, this will work fine if both your host and destination systems are
# using the same c/h/s translation.
#
let asize=$cylinders*$sectorscylinder
#
# Here we remove additional cylinders while
# (totalsize <= (cylinders*(sectors/cylinder))) so that
# we have room for the b partition and still lie on a cylinder boundary:
#
while [ "$totalsize" -le "$asize" ]; do
  let asize=$asize-$sectorscylinder
done
#
# Here we allow ourselves to start after the first track, I guess so
# it can store disklabel or whatever.  Seems like tradition, anyways.
#
# This could break things if sectors per track is different on either
# translation:
#
let asize=$asize-$sectorstrack
#
# BLah.
#
let offset=$asize+$sectorstrack

T=`mktemp /tmp/_ahhXXXXXX`

# Strip out existing partitions
cat $TA | egrep -v ^"  .:" >$T

# Here we setup an 'a' partition that takes up the whole flash media
# and a 'b' partition of minimal size which can be used with mount_mfs
echo "a: $asize	$sectorstrack	4.2BSD	1024	8192	16" >> $T
echo "b: 1		$offset	swap" >> $T
echo "c: $totalsize	0	unused	0	0" >> $T

echo Installing disklabel...
disklabel -R $2 $T

# We don't use $2 in the tar down from here because the user may have
# not specified an absolute path and we change directory for a while...
cp $3 $T

echo
echo Creating new filesystem...
newfs -S $bytessec -u $sectorstrack -z $trackscylinder -q /dev/r$2a
echo

TT=`mktemp -d /tmp/_bhhXXXXXX`

echo Mounting destination to $TT...
if ! mount -o async /dev/$2a $TT; then
  echo Mount failed: mount -o async /dev/$2a $TT
  echo Please fix something.  Aborting...
  echo 
  exit
fi

echo -n Copying OpenBSD distribution to media...
tar cf - -C $5 `cat $T` | tar xpf - -C $TT
echo done

echo Copying bsd kernel, boot blocks, /etc/resolv.conf...
cp $5/usr/mdec/boot $TT/boot
cp /etc/resolv.conf $TT/etc/resolv.conf
cp $4 $TT/bsd

echo Installing boot blocks...
/usr/mdec/installboot -h $trackscylinder -s $sectorstrack $TT/boot $5/usr/mdec/biosboot $2

echo -n Running MAKEDEV...
cdir=`pwd`
mkdir $TT/dev
cd $TT/dev && sh $TT/sbin/MAKEDEV all
echo done
cd $cdir

echo Setting up ttys...
# Since my /etc/rc script leaves /dev read-only, mode 620 is necessary or
# else sshd will not allow anyone to login.
#chgrp tty $TT/dev/ttyp* $TT/dev/ttyq* $TT/dev/ttyr*
chmod 620 $TT/dev/ttyp* $TT/dev/ttyq* $TT/dev/ttyr*

echo Setting up directories and links...
mkdir $TT/tmp $TT/mnt
ln -s /var/run/log $TT/dev/log
ln -s /tmp/var $TT/var

# We are betting here that we will always have /bin/ksh
#
if [ ! -x $TT/bin/sh ]; then
 echo Changing any instance of /bin/sh in $TT/etc/master.passwd to /bin/ksh
 sed -e 's/bin\/sh/bin\/ksh/' <$TT/etc/master.passwd >$TA
 cat <$TA >$TT/etc/master.passwd
fi

if [ ! -x $TT/bin/csh ]; then
 echo Changing any instance of /bin/csh in $TT/etc/master.passwd to /bin/ksh
 sed -e 's/bin\/csh/bin\/ksh/' <$TT/etc/master.passwd >$TA
 cat <$TA >$TT/etc/master.passwd
fi

getpass()
{
 stty -echo
 echo -n "Password:"
 read pw1
 echo
 echo -n "Verify:"
 read pw2
 echo
 stty echo
 if test "$pw1" = "$pw2"; then
  password="$pw1"
 fi
 if test "$password" = ""; then
  echo "Passwords don't match or password was empty.  Try again."
  getpass
 fi
}

if [ "$encpass" == "" ]; then
 if [ "$password" == "" ]; then
  echo
  echo Please assign a root password...
  getpass
  echo
 fi
fi

if [ "$encpass" == "" ]; then
 encpass=`echo $password | /usr/bin/encrypt -b 6`
fi

echo -n "root:$encpass" > $TA

egrep "^root:" < $TT/etc/master.passwd | awk -F : \
  ' { printf(":%s:%s:%s:%s:%s:%s:%s:%s\n", $3, $4, $5, $6, $7, $8, $9, $10) }' \
  >> $TA
egrep -v "^root:" < $TT/etc/master.passwd >> $TA

cat < $TA > $TT/etc/master.passwd

echo "Add user for authpf"
echo -n "Username :"; echo -n ""; read uname
echo -n "Full Name :"; echo -n ""; read user
echo Please assign a password for $uname...
getpass
echo
uid=1001
gid=10
home="/var/empty"
shell="/usr/sbin/authpf"
encpass=`echo $password | /usr/bin/encrypt -b 6`
echo "$uname:$encpass:$uid:$gid::0:0:$user:$home:$shell" >> $TT/etc/master.passwd

pwd_mkdb -p -d $TT/etc $TT/etc/master.passwd

# Adding authpf user to $TT/etc/authpf/authpf.allow
echo $uname >> $TT/etc/authpf/authpf.allow

# Configure more system values ?
# What about configuring the system's syslog server to log everything to
# a central server ? etc...

echo
echo Installation finished.
echo -n Unmounting filesystem...
umount $TT
rmdir $TT
rm $T $TA

echo done!
# bye.......
