#!/bin/sh
# Copyright (C) 2002 Ron Rosson <insane@oneinsane.net> and 
#                    Mike Scott <fugu@oneinsane.net>
#
# Suggestions from Ben Lovett <blovett@users.sourceforge.net>
#
# Build Script for Soekris Appliances using Binaries from installed System
#
# $Id: opensoekris.sh,v 1.33 2004/01/21 07:04:25 blovett Exp $
#

MYNAME=`basename $0`

if [ $# -eq 0 ]; then
    echo "Usage: $MYNAME <build directory>|[-HV]"
    exit 2
fi

if [ "${1}" ]; then
    case $1 in
	-[Hh])
        cat << __EOT
This Script will help you in creating bootable images for your Soekris
hardware. It has its limitations due to that it is only a shell
script. Here is what it is capable of Doing for you:
- Makes an image from the version of OS that you are currently
  building from. ( No Need in grabbing etxXX.tgz or baseXX.tgz
  anymore)
- Function to either right directly to your CF or make an image
  that can be dd'd at a later time
- Customize the configs before the image is created.
- Archive configs, ssh_host keys, and kernels
- And much more.
__EOT
	exit 0 
	;;
    -[Vv])
	echo ""
	echo "Version 1.0.12"
	echo ""
	exit 0 
    	;;
   esac
fi

BUILD=$1
BUILD_HOME=`pwd`
BUILDDIR=$BUILD_HOME/builds/$BUILD
DATE=`date +%Y-%m-%d-%H.%M`

if [ -z "${BSDSRCDIR}" ]; then
   BSDSRCDIR=/usr/src
else
   echo "Using ${BSDSRCDIR} as BSDSRCDIR"
fi

# load functions required
. $BUILD_HOME/configure

# make sure builds dir exists
if [ ! -d $BUILDDIR ]; then
    mkdir -p $BUILD_HOME/builds
fi

if [ -d $BUILDDIR ]; then
    ask_until "$BUILD already exists. Do you want to use it?" n
    DIR=$resp
    case ${DIR} in
        y*|Y*)
           echo "Continuing with the old build dir. You have been warned!"
           ;;
  	*)
	   mv $BUILD_HOME/builds/$BUILD $BUILD_HOME/builds/$BUILD.$DATE
	   echo "Archived BUILD to $BUILDDIR.$DATE"
	   ;;
    esac
fi

# Which editor to use.. Currently only allows vi or mg
editor() {
   ask_until "Which editor do you want to use during the build (vi or mg)?" vi
   case $resp in
   vi)
      export EDITOR=$resp
      ;;
   mg)
      export EDITOR=$resp
      ;;
   *)
      echo -n "Sorry, only mg and vi are allowed at this time. Please chose one"
      echo " of them."
      editor
      ;;
   esac
}

if [ -z "${EDITOR}" ]; then
   editor
fi


if [ ! -d $BUILDDIR ] || [[ `echo $DIR | tr [a-z] [A-Z]` == N ]]; then 
	echo "Creating Directory Structure"
    	mkdir $BUILDDIR
	mkdir $BUILDDIR/root
	mkdir $BUILDDIR/sbin
	mkdir $BUILDDIR/bin
	mkdir $BUILDDIR/etc
	mkdir $BUILDDIR/etc/ssh
	mkdir $BUILDDIR/etc/authpf
	mkdir $BUILDDIR/etc/authpf/banned
	mkdir $BUILDDIR/usr
	mkdir $BUILDDIR/usr/lib
	mkdir $BUILDDIR/usr/bin
	mkdir $BUILDDIR/usr/sbin
	mkdir $BUILDDIR/usr/libexec
	mkdir $BUILDDIR/usr/libexec/auth
	mkdir $BUILDDIR/usr/share
	mkdir $BUILDDIR/usr/share/misc
	mkdir $BUILDDIR/dev

        # change permissions for some dirs
        chown root:auth $BUILDDIR/usr/libexec/auth
        chmod 750 $BUILDDIR/usr/libexec/auth

	echo "Copying Necessary Files Over"

	# Copy /etc files from ${BSDSRCDIR}/etc
	cd ${BSDSRCDIR}/etc
	SRCDIR=/tmp/temproot-$$
	mkdir $SRCDIR
	make DESTDIR=$SRCDIR distribution-etc-root-var >/dev/null
	cd $BUILD_HOME

SRCBINFILES="root/.profile root/.klogin"

SBINFILES="brconfig init mount ifconfig umount ping ping6 reboot \
           route dmesg fsck_ffs fsck mount_mfs mount_ffs mknod \
           ldconfig pfctl pflogd sysctl wicontrol"

USRSBINFILES="arp authpf sshd dev_mkdb pwd_mkdb tcpdump traceroute \
              traceroute6 vipw"

USRBINFILES="login netstat touch du grep scp ssh ssh-keygen nc passwd \
             top less logger"

BINFILES="cat sync pwd ps mv ls ln ksh hostname echo df date cp \
          chmod rm sh kill stty mkdir sleep"

SRCETCFILES="protocols services moduli login.conf master.passwd \
             gettytab passwd group spwd.db pwd.db pf.os"

LIBEXECFILES="ld.so getty"

LIBEXECAUTHFILES="login_passwd"

MISCFILES="/usr/share/misc/terminfo.db"

DEVFILES="MAKEDEV"

BUILDETCFILES="authpf/authpf.allow authpf/authpf.rules authpf/authpf.nat \
               authpf/authpf.message authpf/authpf.problem"
	       
	for file in ${SRCBINFILES}; do
		cp -p $SRCDIR/${file} $BUILDDIR/root/
	done

	for file in ${SBINFILES}; do
		cp -p /sbin/${file} $BUILDDIR/sbin/
	done

	for file in ${USRSBINFILES}; do
		cp -p /usr/sbin/${file} $BUILDDIR/usr/sbin/
	done

	for file in ${BINFILES}; do
		cp -p /bin/${file} $BUILDDIR/bin/
	done

	for file in ${USRBINFILES}; do
		cp -p /usr/bin/${file} $BUILDDIR/usr/bin/
	done

	for file in ${SRCETCFILES}; do
		cp -p $SRCDIR/etc/${file} $BUILDDIR/etc/
	done

        # copy ssh*_config's manually for now
        cp $SRCDIR/etc/ssh/ssh_config $BUILDDIR/etc/ssh
        cp $SRCDIR/etc/ssh/sshd_config $BUILDDIR/etc/ssh

	for file in ${LIBEXECFILES}; do
		cp -p /usr/libexec/${file} $BUILDDIR/usr/libexec/
	done

        for file in ${LIBEXECAUTHFILES}; do
            cp -p /usr/libexec/auth/${file} ${BUILDDIR}/usr/libexec/auth
        done

	for file in ${MISCFILES}; do
		cp -p ${file} $BUILDDIR/usr/share/misc/
	done

	for file in ${DEVFILES}; do
		cp -p /dev/${file} $BUILDDIR/sbin/
	done

	cp -pr /usr/mdec $BUILDDIR/usr/
	cp $BUILD_HOME/ttys $BUILDDIR/etc/
	cp $BUILD_HOME/fstab $BUILDDIR/etc/
	cp $BUILD_HOME/boot.conf $BUILDDIR/etc/

	# files we need to touch
	echo "Touching other files..."

	for file in ${BUILDETCFILES}; do
		touch $BUILDDIR/etc/$file
	done

	ln ${BUILDDIR}/bin/chmod ${BUILDDIR}/bin/chgrp
	ln ${BUILDDIR}/bin/chmod ${BUILDDIR}/sbin/chown
	ln ${BUILDDIR}/sbin/reboot ${BUILDDIR}/sbin/halt

        # cleanup $SRCDIR
        rm -rf $SRCDIR
fi

cp $BUILD_HOME/flashsmall.txt $BUILD_HOME/flashsmall.tmp

# Deal with pf.conf
ask "Do you have a pf.conf you want to use?" n
PF=$resp
case $PF in
  y*|Y*)
     ask "Filename?" ""
     if [ -f ${resp} ]; then
	cp ${resp} $BUILDDIR/etc/pf.conf
     fi
     ;;
  *)
     cp -p ${BSDSRCDIR}/etc/pf.conf $BUILDDIR/etc/
     ask "Do you want to edit the stock pf.conf?" y
     case $resp in
	y*|Y*)
	    $EDITOR $BUILDDIR/etc/pf.conf
	    ;;
        *)
            ;;
     esac
     ;;
esac

# Deal with authpf.conf
ask "Do you have an authpf.conf you want to use?" n
HAVEAUTHPFCONF=$resp
case $HAVEAUTHPFCONF in
 n*|N*)
     cat << __EOF > $BUILDDIR/etc/authpf/authpf.conf
anchor=authpf
__EOF
     ;;
 *)
     ask "Enter filename: " ""
     AUTHPFCONFFILE=$resp
     if [ -f $AUTHPFCONFFILE ]; then
        cp -p $AUTHPFCONFFILE $BUILDDIR/etc/authpf/authpf.conf
     fi
     ;;
esac 

# Deal with authpf.rules
ask "Do you have an authpf.rules you want to use?" n
HAVEAUTHPFRULES=$resp
case $HAVEAUTHPFRULES in
  n*|N*)
     echo "Building basic authpf.rule"
     cat << __EOF > $BUILDDIR/etc/authpf/authpf.rules
# Basic authpf.rules file
internal_if = "wi0"
pass in quick on \$internal_if proto { tcp, udp, icmp } from \$user_ip/32 to any keep state label \$user_id
__EOF

     ask "Do you want to edit the authpf.rules?" y
     EDITAUTHPFRULES=$resp
     case $EDITAUTHPFRULES in
       y*|Y*)
	    $EDITOR $BUILDDIR/etc/authpf/authpf.rules	
            ;;
           *)
            ;;
     esac
     ;;
  *)
     ask_until "Enter filename: " ""
     AUTHPFRULESFILE=$resp
     if [ -f $AUTHPFRULESFILE ]; then
        cp $AUTHPFRULESFILE $BUILDDIR/etc/authpf/authpf.rules
     else
        echo "The file (${AUTHPFRULESFILE}) did not exist. You'll have to fix this later."
     fi
     ;;
esac

# Change sshd_config for better security with authpf
 echo "Tweaking sshd_config for authpf"
 cat << __EOF >> $BUILDDIR/etc/ssh/sshd_config
ClientAliveInterval 15
ClientAliveCountMax 3
__EOF

# Deal with ssh keys
 if [ -f $BUILDDIR/etc/ssh/ssh_host_dsa_key ]; then
     ask "SSH keys already exist, do you want to keep them?" n
     SSHKEYSEXIST=$resp
     case $SSHKEYSEXIST in 
         n*|N*)
             rm $BUILDDIR/etc/ssh/ssh_host_dsa_key*
             rm $BUILDDIR/etc/ssh/ssh_host_rsa_key*
             rm $BUILDDIR/etc/ssh/ssh_host_key*
             ask "SSH keys removed. Do you have some new SSH host keys that you want to use? " y
             HAVESSHKEYS=$resp
             case $HAVESSHKEYS in
                 y*|Y*)
                     ask_until "Enter directory: " ""
                     SSHKEYSDIR=$resp
		     if [ -d $SSHKEYSDIR ]; then
                     	cp $SSHKEYSDIR/ssh_host* $BUILDDIR/etc/ssh/
		     else
                        echo "The directory $SSHKEYSDIR does not exist. Please fix this after the build is done..."
                     fi
                     ;;
         	 n*|N*)
	     	     echo "Generating RSA, DSA, and RSA1 keys..."
             	     /usr/bin/ssh-keygen -q -t dsa -f $BUILDDIR/etc/ssh/ssh_host_dsa_key -N ''
             	     /usr/bin/ssh-keygen -q -t rsa -f $BUILDDIR/etc/ssh/ssh_host_rsa_key -N ''
             	     /usr/bin/ssh-keygen -q -t rsa1 -f $BUILDDIR/etc/ssh/ssh_host_key -N ''
                     ;;
	     esac
     esac
 else 			# keys don't exit in BUILDDIR
     ask "Do you have a ssh host keys you want to use? " n
     HAVESSHKEYS=$resp
     case $HAVESSHKEYS in
        n*|N*)
           echo "Generating DSA, RSA, and RSA1 keys."
           /usr/bin/ssh-keygen -q -t dsa -f $BUILDDIR/etc/ssh/ssh_host_dsa_key -N ''
           /usr/bin/ssh-keygen -q -t rsa -f $BUILDDIR/etc/ssh/ssh_host_rsa_key -N ''
           /usr/bin/ssh-keygen -q -t rsa1 -f $BUILDDIR/etc/ssh/ssh_host_key -N ''
	   ;;
           *)
           ask_until "Enter directory" ""
           SSHKEYSDIR=$resp
           if [ -d ${SSHKEYSDIR} -a -f ${SSHKEYSDIR}/ssh_host_rsa_key ]; then
              cp $SSHKEYSDIR/ssh_host* $BUILDDIR/etc/ssh/
           else
              echo "Sorry. I can't find any keys in ${SSHKEYSDIR}. Building new ones"
              /usr/bin/ssh-keygen -q -t dsa -f $BUILDDIR/etc/ssh/ssh_host_dsa_key -N ''
              /usr/bin/ssh-keygen -q -t rsa -f $BUILDDIR/etc/ssh/ssh_host_rsa_key -N ''
              /usr/bin/ssh-keygen -q -t rsa1 -f $BUILDDIR/etc/ssh/ssh_host_key -N ''
	   fi
 	   ;;
     esac
 fi

# Check for the existance of a rc script
 ask "Do you have a rc script that you want to use?" n
 HAVERC=$resp
 case $HAVERC in
   [Nn])
      configure
      cp $BUILD_HOME/rc $BUILDDIR/etc
      cat $BUILD_HOME/rc.tmp >> $BUILDDIR/etc/rc
      rm $BUILD_HOME/rc.tmp
      ;;
   *)
     ask_until "Enter the location of the file: " ""
     RCLOC=$resp
     if [ -f $RCLOC ]; then
         cp $RCLOC $BUILDDIR/etc/rc
     else
         echo "Ack. That file doesn't exist. Going ahead with a interactive build"
         echo " of the rc script.."
         configure
         cp $BUILD_HOME/rc $BUILDDIR/etc
         cat $BUILD_HOME/rc.tmp > $BUILDDIR/etc/rc
         rm $BUILD_HOME/rc.tmp
     fi
     ;;
esac

# configure timezone
 ask "Would you like the timezone to be configured for you? (Defaults to UTC)" y
 CONFIGTZ=$resp
case $CONFIGTZ in
   y*|Y*)
     cp /etc/localtime $BUILDDIR/etc/localtime
     ;;
    *)
     ;;
esac
 
# Wanna save your work for later?
BACKUPFILES="rc pf.conf authpf/authpf.conf authpf/authpf.rules ssh/ssh_host*"

 ask "Do you want to save these configs/keys for posterity?" y
 BACKUPCONFIGS=$resp
case $BACKUPCONFIGS in
   y*|Y*)
     echo "Saving to $BUILD_HOME/conf/$BUILD"
     if [ ! -d $BUILD_HOME/conf/$BUILD ]; then
	 mkdir -p $BUILD_HOME/conf/$BUILD
     fi

     for file in ${BACKUPFILES}; do
     	cp $BUILDDIR/etc/$file $BUILD_HOME/conf/$BUILD/
     done
     ;;
   *)
     ;;
esac

# Check for existing kernel
 echo "Checking for existing kernel"
 if [ ! -e  $BUILDDIR/bsd ]; then
     if [ ! -e $BUILD_HOME/conf/bsd ]; then
         ask "Do you have a kernel already built?" n
	 case ${resp} in
	 [Nn])
             case $HAVERC in
               n*|N*)
                 if [ ! -e ${BSDSRCDIR}/sys/arch/i386/conf/NET${DEVICE} ]; then
                     cp ${BUILD_HOME}/NET${DEVICE} ${BSDSRCDIR}/sys/arch/i386/conf/
                 fi
                 buildkernel $DEVICE
               ;;
                *)
                 ask_until "What Device are we building this kernel for (NET4501|NET4511|NET4521|NET4801)?" ""
                 DEVICE=`echo $resp | cut -c4 -c5 -c6 -c7`
                 if [ ! -e ${BSDSRCDIR}/sys/arch/i386/conf/NET${DEVICE} ]; then
                     cp ${BUILD_HOME}/NET${DEVICE} ${BSDSRCDIR}/sys/arch/i386/conf/
                 fi
                 buildkernel $DEVICE
               ;;
             esac
	     ;;
	    *)
                 ask_until "Enter a filename: " ""
                 KERNLOC=$resp
	         if [ -e $KERNLOC ]; then
		    cp $KERNLOC $BUILDDIR/bsd
                    case ${BACKUPCONFIGS} in
  	            [Yy])
		       cp $KERNLOC $BUILD_HOME/conf/$BUILD/bsd
		       echo "Saved this kernel to $BUILD_HOME/conf/$BUILD/ also"
	 	       ;;
		    esac
	         else
		    echo "You fubar'ed something, building a kernel anyway!"
		    ask_until "What Device are we building this kernel for (NET4501|NET4511|NET4521|NET4801)?" ""
                    DEVICE=`echo $DEVICE | cut -c4 -c5 -c6 -c7`
		    if [ ! -e ${BSDSRCDIR}/sys/arch/i386/conf/NET${DEVICE} ]; then
			cp ${BUILD_HOME}/NET${DEVICE} ${BSDSRCDIR}/sys/arch/i386/conf/
		    fi
		    buildkernel $DEVICE
		 fi
	     ;;
	 esac
     else
	 cp $BUILD_HOME/conf/bsd $BUILDDIR/bsd
     fi
 fi

# Check to see if pftop is installed (through ports...) on build system
ask "Do you want to install pftop (requires a ports tree)" n
INSTALLPFTOP=$resp
case $INSTALLPFTOP in
n*|N*)
   ;;
*)
   echo  "Installing PFTOP"
    if [ -x /usr/local/sbin/pftop ]; then
       cp -p /usr/local/sbin/pftop $BUILDDIR/sbin
     else
       WRKOBJDIR=`grep WRKOBJDIR /etc/mk.conf | cut -d ' ' -f 2`
       PFTOPVER=`grep DISTNAME /usr/ports/sysutils/pftop/Makefile | cut -f 2`
       if [ -z ${WRKOBJDIR} ]; then
         cd /usr/ports/sysutils/pftop && make && cp w-${PFTOPVER}/${PFTOPVER}/pftop ${BUILDDIR}/sbin
         echo "pftop built and installed in $BUILDDIR..."
       else 
         cd /usr/ports/sysutils/pftop && make && cp ${WRKOBJDIR}/${PFTOPVER}/${PFTOPVER}/pftop ${BUILDDIR}/sbin
         echo "pftop built and installed in $BUILDDIR..."
       fi
    fi

   echo "./sbin/pftop" >> ${BUILD_HOME}/flashsmall.tmp
   ;;
esac

# Build and install nsh
ask "Do you want to install nsh?" y
INSTALLNSH=$resp
case $INSTALLNSH in
n*|N*)
   ;;
*)
   echo "Installing NSH"
   if [ -d $BUILD_HOME/nsh ]; then
      cd $BUILD_HOME/nsh
      make
      cp $BUILD_HOME/nsh/nsh $BUILDDIR/sbin 
      make clean
      echo "./sbin/nsh" >> $BUILD_HOME/flashsmall.tmp
   fi
   ;;
esac

# Which editor?
ask_until "Which editor do you want, vi or mg (emacs like)?" vi
INSTEDITOR=$resp
case $INSTEDITOR in
vi)
   cp -p /usr/bin/vi ${BUILDDIR}/usr/bin
   echo "./usr/bin/vi" >> $BUILD_HOME/flashsmall.tmp
   ;;
mg)
   cp -p /usr/bin/mg ${BUILDDIR}/usr/bin
   echo "./usr/bin/mg" >> $BUILD_HOME/flashsmall.tmp
   ;;
*)
   echo "Unknown answer, defaulting to vi"
   cp -p /usr/bin/vi ${BUILDDIR}/usr/bin
   echo "./usr/bin/vi" >> $BUILD_HOME/flashsmall.tmp
   ;;
esac
  
# Adding Private file to the Tree (Addons)
cp $BUILD_HOME/addon/telnet $BUILDDIR/usr/bin
 
	 
# Writing The Media
cd $BUILD_HOME

# Find out what size CF we are using
ask_until "What size CF are we going to use (megs)" ""
CFSIZE=$resp
case $CFSIZE in
    32)
	COUNT=62720
	;;
    64)
	COUNT=124928
	;;
    128)
	COUNT=250368
	;;
    256)
	COUNT=501760
	;;
    1024)
	COUNT=2104704
	;;
    *)
	echo "Please edit this file and the flashdist.sh and send"
	echo "the diffs to the OpenSoekris Team."
	exit 0
	;;
esac

# Dynamically generate list of required libraries
echo "Dynamically generating list of required libraries..."
cd $BUILDDIR
cat $BUILD_HOME/flashsmall.tmp | xargs -n1 ldd 2> /dev/null | sed -ne's|.* \(/[^ ]*\.so[^ ]*\).*|\1|gp' | sort | uniq > $BUILD_HOME/libs.tmp
for lib in `cat $BUILD_HOME/libs.tmp`; do
   if [ $lib != "/usr/libexec/ld.so" ]; then
      echo ".$lib" >> $BUILD_HOME/flashsmall.tmp
      cp -p $lib $BUILDDIR/$lib
   fi
done
cd $BUILD_HOME
rm libs.tmp

ask "Do you want to write this to an image or CF [I/C]?" i
WRITETO=$resp
case $WRITETO in
   i*|I*)
      echo -n "Creating the image .."
      # make sure image directory exists

      if [ ! -d $BUILD_HOME/image ]; then
         echo "Making $BUILD_HOME/image directory because it didn't exist..."
         mkdir $BUILD_HOME/image
      fi

      dd if=/dev/zero of=image/$BUILD.$DATE.bin bs=512 count=$COUNT
      vnconfig svnd0 image/$BUILD.$DATE.bin
      ./flashdist.sh $CFSIZE svnd0 flashsmall.tmp $BUILDDIR/bsd $BUILDDIR
      vnconfig -u svnd0
      rm flashsmall.tmp
   ;;
   c*|C*)
      writecf
   ;;
esac
