/* $Id: pflogaccept.c,v 1.5 2003/03/18 05:02:35 jnazario Exp $ */

/*
 * Copyright (c) 2003 Jose Nazario <jose@monkey.org>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <arpa/inet.h>

#include <err.h>
#include <pcap.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static void 
usage (char *progname) {
	printf ("%s: accepts pflog messages sent by pflogexport\n", progname);
	printf ("usage: %s -p port -d device\n", progname);
	printf ("\tp\tlistening port (default 9999)\n");
	printf ("\td\toutput device (default pflog0)\n");
	exit (0);
}

int 
main (int argc, char *argv[]) {
	int sockfd, bindno, c, lport, m, n;
	char *outdev, msg[2048], errbuf[PCAP_ERRBUF_SIZE];
	struct sockaddr_in serv_addr;
	fd_set fds;
	pcap_t *pfl;
	struct timeval to;

	lport = 9999;
	outdev = "pflog0";

	while ((c = getopt(argc, argv, "p:d:h?")) != -1) {
		switch (c) {
		case 'p':
			lport = atoi(optarg);
			break;
		case 'd':
			outdev = optarg;
			break;
		default:
			usage(argv[0]);
			break;
		}
	}

	if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		printf ("socket error\n");
		exit (-1);
	}
	
	bzero ((char *)&serv_addr, sizeof (serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	serv_addr.sin_port = htons(lport);

	if ((bindno = bind(sockfd, (struct sockaddr *) &serv_addr, 
			   sizeof(serv_addr))) < 0) {
		printf("server error: can't bind local address\n");
		exit(2);
	}


	if ((pfl = pcap_open_live (outdev, 1540, 0, 500, errbuf)) == NULL)
		printf ("unable to open %s device: %s\n", outdev, errbuf);

	while (1) {
		FD_ZERO (&fds);
		FD_SET (sockfd, &fds);
		to.tv_sec = 1;
		to.tv_usec = 0;

		m = select (sockfd + 1, &fds, NULL, NULL, &to);
		if (m < 0) {
			printf ("error: select\n");
			break;
		}
		if (m > 0) {
			if (FD_ISSET (sockfd, &fds)) {
				ioctl (sockfd, FIONBIO, 1);  
				n = read (sockfd, msg, 2048);
				printf("%% Message of size %d\n", n);
				//pcap_inject (pfl, (void *)msg, sizeof(msg));
				write (pcap_fileno(pfl), msg, n);
				FD_CLR (sockfd, &fds);
			}
		}
	}

	close (sockfd);
	pcap_close (pfl);
	exit (0);
}
