/* $Id: pflogexport.c,v 1.3 2003/03/19 03:44:51 jnazario Exp $ */

/*
 * Copyright (c) 2003 Jose Nazario <jose@monkey.org>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <arpa/inet.h>

#include <err.h>
#include <pcap.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

struct myctx {
	struct in_addr	 dst;		/* destination host */
	u_int		 dport;		/* destination port */
};

struct pfm {
	struct pcap_pkthdr *pkth;
	u_char pkt[1500];
};

pcap_t         *p = NULL;
int		u;

static void
usage (const char *prog) {
	printf ("usage: %s -d addr -p port\n", prog);
	exit (0);
}

static void
pf_add (u_char *args, const struct pcap_pkthdr *pkth, const u_char *pkt) {
	struct pfm pfpayload;

	memmove (&pfpayload.pkth, pkth, sizeof(pkth));
	memmove (&pfpayload.pkt, pkt, pkth->caplen);
	write (u, &pfpayload, sizeof(pkth) + pkth->caplen);
}

int
main (int argc, char *argv[]) {
	int c, on;
	char *intf, errbuf[PCAP_ERRBUF_SIZE];
	struct myctx ctx;
	struct sockaddr_in srv;

	inet_pton (AF_INET, "127.0.0.1", &ctx.dst);
	ctx.dport = 9999;
	intf = "pflog0";

	while ((c = getopt(argc, argv, "d:f:i:p:h?")) != -1) {
		switch (c) {
		case 'd':
			if (inet_pton (AF_INET, optarg, &ctx.dst) != 1) {
				printf ("invalid destination address\n");
				exit (-1);
			}
			break;
		case 'i':
			intf = optarg;
			break;
		case 'p':
			ctx.dport = atoi(optarg);
			break;
		default:
			usage(argv[0]);
		}
	}

	if ((p = pcap_open_live (intf, 1540, 0, 100, errbuf)) == NULL) {
		printf ("pcap open error: %s\n", errbuf);
		exit (-1);
	}
	on = 1;
	if (ioctl(pcap_fileno(p), BIOCIMMEDIATE, &on) < 0)
		warn("BIOCIMMEDIATE");

	bzero (&srv, sizeof (srv));
	srv.sin_family = AF_INET;
	srv.sin_addr = ctx.dst;
	srv.sin_port = htons(ctx.dport);
	if ((u = socket (AF_INET, SOCK_DGRAM, 0)) < 0) {
		printf ("socket error\n");
		exit (-1);
	}
	connect (u, (struct sockaddr *)&srv, sizeof(srv));

	daemon (0, 0); 

	for (;;) {
		pcap_loop(p, 0, pf_add, NULL);
	}

	pcap_close (p);
	close (u);
	return (1);
}
